import os
from executor import execute
from .invoke import DEFAULT_QUEUE_NAME


class Config(object):
    def __init__(
        self,
        local_repository=None,
        aws_region="us-east-1",
        ecr_repository=None,
        ecs_task_name="werkit-default",
        task_args=None,
        default_task_count=1,
        redis_url=None,
        queue_name=DEFAULT_QUEUE_NAME,
    ):
        self.local_repository = local_repository
        self.aws_region = aws_region
        self.ecr_repository = ecr_repository
        self.ecs_task_name = ecs_task_name
        self.task_args = task_args
        self.default_task_count = default_task_count
        self.redis_url = redis_url or os.environ.get("WERKIT_REDIS_URL", None)
        self.queue_name = queue_name


class CloudManager(object):
    def __init__(self, config):
        if not isinstance(config, Config):
            raise ValueError("Should be an instance of Config")
        self.config = config

    @property
    def redis_connection(self):
        from redis import Redis

        if not self.config.redis_url:
            raise ValueError("Redis URL must be configured")

        return Redis.from_url(self.config.redis_url)

    def login(self):
        login_cmd = execute(
            "aws",
            "ecr",
            "get-login",
            "--region",
            self.config.aws_region,
            "--no-include-email",
            capture=True,
        )
        execute(login_cmd)

    def remote_tag(self, tag):
        return "{}:{}".format(self.config.ecr_repository, tag)

    def build(self, tag, extra_build_args=[]):
        local_tag = "{}:{}".format(self.config.local_repository, tag)
        execute(
            "docker",
            "build",
            "-f",
            "Dockerfile-rq.dev",
            ".",
            "--tag",
            local_tag,
            *extra_build_args
        )
        execute("docker", "tag", local_tag, self.remote_tag(tag))

    def build_and_push(self, tag, extra_build_args=[]):
        self.build(tag=tag, extra_build_args=extra_build_args)
        execute("docker", "push", self.remote_tag(tag))

    def clean(self):
        from .invoke import clean as _clean

        _clean(queue_name=self.config.queue_name, connection=self.redis_connection)

    def run(self, tag, count=None):
        if count is None:
            count = self.config.default_task_count
        execute(
            "fargate",
            "task",
            "run",
            self.config.ecs_task_name,
            "--image",
            self.remote_tag(tag),
            "--num",
            str(count),
            "--env",
            "REDIS_URL={}".format(self.config.redis_url),
            *self.config.task_args
        )

    def dashboard(self):
        execute(
            "rq-dashboard", "--bind", "localhost", "--redis-url", self.config.redis_url
        )

    def ps(self):
        execute("fargate", "task", "ps", self.config.ecs_task_name)

    def get_results(self, wait_until_done=False):
        from .invoke import get_results as _get_results

        return _get_results(
            wait_until_done=wait_until_done,
            queue_name=self.config.queue_name,
            connection=self.redis_connection,
        )



import asyncio
import concurrent
import datetime
import os
from botocore.exceptions import ClientError
from harrison import Timer
from .parallel import parallel_map_on_lambda
from ..schema import validate_result, wrap_exception

LAMBDA_WORKER_FUNCTION_NAME = "LAMBDA_WORKER_FUNCTION_NAME"
env_lambda_worker_function_name = os.environ.get(LAMBDA_WORKER_FUNCTION_NAME)

LAMBDA_WORKER_TIMEOUT = "LAMBDA_WORKER_TIMEOUT"
env_lambda_worker_timeout = (
    int(os.environ[LAMBDA_WORKER_TIMEOUT])
    if LAMBDA_WORKER_TIMEOUT in os.environ
    else None
)


def transform_result(result, start_time):
    if (
        isinstance(result, ClientError)
        or isinstance(result, asyncio.TimeoutError)
        # TODO: Could this be replaced with just the following clause?
        or isinstance(result, Exception)
    ):
        return wrap_exception(
            exception=result, error_origin="orchestration", start_time=start_time
        )
    elif isinstance(result, dict) and "errorMessage" in result:
        # https://docs.aws.amazon.com/lambda/latest/dg/python-exceptions.html
        # Unhandled exception in Lambda, with `errorMessage`, `errorType`, and
        # `stackTrace`.
        return {
            "success": False,
            "result": None,
            "error": result["stackTrace"]
            + [f"{result['errorType']}: {result['errorMessage']}"],
            "error_origin": "system",
            "start_time": start_time.astimezone().isoformat(),
            "duration_seconds": -1,
        }
    else:
        validate_result(result)
        return result


def handler(
    event,
    context,
    lambda_worker_function_name=env_lambda_worker_function_name,
    timeout=env_lambda_worker_timeout or 120,
):
    start_time = datetime.datetime.now()
    start_timestamp = datetime.datetime.utcnow().timestamp()

    with Timer(verbose=False) as response_timer:
        if not lambda_worker_function_name:
            raise Exception(
                f"Environment variable {LAMBDA_WORKER_FUNCTION_NAME} must be defined, "
                + "or default kwArg lambda_worker_function_name must be bound to the handler"
            )
        event_loop = asyncio.get_event_loop()
        with concurrent.futures.ThreadPoolExecutor(max_workers=1023) as executor:
            results = event_loop.run_until_complete(
                parallel_map_on_lambda(
                    lambda_worker_function_name,
                    timeout,
                    event_loop=event_loop,
                    executor=executor,
                    **event,
                )
            )
            results = [
                transform_result(result=result, start_time=start_time)
                for result in results
            ]
    return {
        "results": results,
        "orchestrator_duration_seconds": response_timer.elapsed_time_s,
        "start_timestamp": start_timestamp,
    }
